"""Defines the metalens challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Tuple

from fmmax import fmm  # type: ignore[import-untyped]
import jax
from jax import nn
from jax import numpy as jnp
from totypes import symmetry, types

from invrs_gym.challenges import base
from invrs_gym.challenges.metalens import component as metalens_component
from invrs_gym.utils import initializers


ENHANCEMENT_EX_MEAN = "enhancement_ex_mean"
ENHANCEMENT_EX_MIN = "enhancement_ex_min"
ENHANCEMENT_EY_MEAN = "enhancement_ey_mean"
ENHANCEMENT_EY_MIN = "enhancement_ey_min"

EX = "ex"
EY = "ey"


density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class MetalensChallenge(base.Challenge):
    """Defines the metalens challenge.

    The challenge is based on the RGB metalens problem from "Validation and
    characterization of algorithms and software for photonics inverse design" by
    Chen et al. It involves optimization of a metalens to focus 450, 550, and 650 nm
    light a distance 2.4 um away from a 1D metalens having a width of 10 um.

    https://opg.optica.org/josab/ViewMedia.cfm?uri=josab-41-2-A161
    https://github.com/NanoComp/photonics-opt-testbed/tree/main/RGB_metalens

    Attributes:
        component: The component to be designed.
        incident_field: Either `EX` or `EY`, specifying whether the target of the
            challenge is to optimize for excitation with x-polarized or y-polarized
            electric fields.
    """

    component: metalens_component.MetalensComponent
    incident_field: str

    def loss(self, response: metalens_component.MetalensResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        assert self.incident_field in (EX, EY)
        if self.incident_field == EX:
            enhancement = response.enhancement_ex
        else:
            enhancement = response.enhancement_ey

        return soft_amax(-enhancement, scale=10.0)

    def eval_metric(
        self,
        response: metalens_component.MetalensResponse,
    ) -> jnp.ndarray:
        """Computes the eval metric from the component `response`.

        The evaluation metric is the intensity enhancement for the wavelength having
        minimum intensity at the target point.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        assert self.incident_field in (EX, EY)
        if self.incident_field == EX:
            enhancement = response.enhancement_ex
        else:
            enhancement = response.enhancement_ey
        return jnp.amin(enhancement)

    def metrics(
        self,
        response: metalens_component.MetalensResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Args:
            response: The response of the metalens component.
            params: The parameters where the metalens was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary.
        """
        metrics = super().metrics(response, params, aux)
        metrics.update(
            {
                ENHANCEMENT_EX_MIN: jnp.amin(response.enhancement_ex),
                ENHANCEMENT_EX_MEAN: jnp.mean(response.enhancement_ex),
                ENHANCEMENT_EY_MIN: jnp.amin(response.enhancement_ey),
                ENHANCEMENT_EY_MEAN: jnp.mean(response.enhancement_ey),
            }
        )
        return metrics


def soft_amax(x: jnp.ndarray, scale: float) -> jnp.ndarray:
    """A soft version of `amax`.

    The softness is set by `scale`. For small values, the output is close to that of
    `amax`, while for larger values it is closer to that of `mean`. This function can
    be used to scalarize a vector objective in a manner related to the concept of
    minimax optimization.

    Args:
        x: The array to be scalarized.
        scale: The scale of smoothness.

    Returns:
        The scalarized array.
    """
    return jnp.sum(jax.lax.stop_gradient(nn.softmax(x / scale)) * x)


METALENS_SPEC = metalens_component.MetalensSpec(
    permittivity_ambient=(1.0 + 0.0001j) ** 2,
    permittivity_metalens=(2.4 + 0.0001j) ** 2,
    permittivity_substrate=(2.4 + 0.0001j) ** 2,
    thickness_ambient=1.0,  # Above the focal point
    thickness_lens=1.0,
    thickness_substrate=1.0,
    width_lens=10.0,
    width_pml=0.0,
    pml_lens_offset=3.0,
    pml_source_offset=2.5,
    source_smoothing_fwhm=1.0,
    focus_offset=2.4,
    grid_spacing=0.02,
)

METALENS_SIM_PARAMS = metalens_component.MetalensSimParams(
    wavelength=jnp.asarray([0.45, 0.55, 0.65]),
    approximate_num_terms=280,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    num_layers=25,
)

SYMMETRIES: Tuple[str, ...] = (symmetry.REFLECTION_N_S,)

# Minimum width and spacing are 100 nm for the default dimensions.
MINIMUM_WIDTH = 5
MINIMUM_SPACING = 5

INTENSITY_ENHANCEMENT_LOWER_BOUND = 15.0


def metalens(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
    incident_field: str = EX,
    spec: metalens_component.MetalensSpec = METALENS_SPEC,
    sim_params: metalens_component.MetalensSimParams = METALENS_SIM_PARAMS,
    symmetries: Tuple[str, ...] = SYMMETRIES,
) -> MetalensChallenge:
    """Metalens extractor with 10.0 x 1.0 um design region.

    The challenge is based on the RGB metalens problem from "Validation and
    characterization of algorithms and software for photonics inverse design" by
    Chen et al. It involves optimization of a metalens to focus 450, 550, and 650 nm
    light a distance 2.4 um away from a 1D metalens having a width of 10 um.

    https://opg.optica.org/josab/ViewMedia.cfm?uri=josab-41-2-A161
    https://github.com/NanoComp/photonics-opt-testbed/tree/main/RGB_metalens

    Args:
        minimum_width: The minimum width target for the challenge, in pixels.  The
            default value of 5 corresponds to a physical size of approximately 100 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        incident_field: Either `EX` or `EY`, specifying whether the target of the
            challenge is to optimize for excitation with x-polarized or y-polarized
            electric fields.
        spec: Defines the physical specification of the metalens.
        sim_params: Defines the simulation settings of the metalens.
        symmetries: Defines the symmetries of the metalens.

    Returns:
        The `MetalensChallenge`.
    """
    return MetalensChallenge(
        component=metalens_component.MetalensComponent(
            spec=spec,
            sim_params=sim_params,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        incident_field=incident_field,
    )



"""Defines the photon extractor component and simulation routine.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Any, Optional, Tuple

import jax
import jax.numpy as jnp
from fmmax import basis, fields, fmm, pml, scattering, sources  # type: ignore[import]
from jax import tree_util
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0

EFIELD = "efield"
HFIELD = "hfield"
FIELD_COORDINATES = "field_coordinates"


@dataclasses.dataclass
class ExtractorSpec:
    """Defines the physical specifcation of a photon extractor.

    Args:
        permittivity_ambient: Permittivity of the ambient material.
        permittivity_oxide: Permittivity of the oxide material.
        permittivity_extractor: Permittivity of the extractor material.
        permittivity_substrate: Permittivity of the substrate.
        thickness_ambient: The thickness of the ambient layer.
        thickness_oxide: The thickness of the oxide layer.
        thickness_extractor: The thickness of the extractor layer.
        thickness_substrate_before_source: The distance between the substrate and
            the plane containing the source.
        thickness_substrate_after_source: The thickness of the substrate below the
            source plane.
        width_design_region: The width of the square design region.
        width_padding: Width of the region between the design and the PML.
        width_pml: Width of the perfectly matched layers at the borders of the
            simulation unit cell.
        fwhm_source: The spatial full-width at half maximum for the Gaussian dipole.
        offset_monitor_source: The distance along the z direction between the source
            and monitor planes above and below the source used to compute the total
            power emitted by the source.
        offset_monitor_ambient: The distance along the z direction between the top of
            the extractor structure and a monitor plane used to compute the total
            power extracted from the source.
        width_monitor_ambient: The length on one side of the square flux monitor
            above the design region.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    permittivity_ambient: complex
    permittivity_oxide: complex
    permittivity_extractor: complex
    permittivity_substrate: complex

    thickness_ambient: float
    thickness_oxide: float
    thickness_extractor: float
    thickness_substrate_before_source: float
    thickness_substrate_after_source: float

    width_design_region: float
    width_padding: float
    width_pml: float

    fwhm_source: float

    offset_monitor_source: float
    offset_monitor_ambient: float
    width_monitor_ambient: float

    grid_spacing: float

    @property
    def pitch(self) -> float:
        return self.width_design_region + 2 * (self.width_padding + self.width_pml)

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        return (
            int(jnp.ceil(self.pitch / self.grid_spacing)),
            int(jnp.ceil(self.pitch / self.grid_spacing)),
        )


@dataclasses.dataclass
class ExtractorSimParams:
    """Parameters that configure the simulation of a photon extractor.

    Attributes:
        wavelength: The wavelength of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    wavelength: float | jnp.ndarray
    formulation: fmm.Formulation
    approximate_num_terms: int
    truncation: basis.Truncation


@dataclasses.dataclass
class ExtractorResponse:
    """Contains the response of the photon extractor.

    Attributes:
        wavelength: The wavelength for the efficiency calculation.
        emitted_power: The total power emitted by the source.
        extracted_power: The total power extracted from the source, including power
            at large angles which is not included in `collected_power`.
        collected_power: The total power collected from the source, collected by the
            ambient monitor above the extractor. Since the monitor is smaller than
            the unit cell, not all emitted power is counted as collected.
        bare_substrate_emitted_power: The total power emitted by the source in the
            case where there is no extractor structure.
        bare_substrate_extracted_power: The total power extracted from the source in the
            case where there is no extractor structure.
        bare_substrate_collected_power: The total power collected from the source in the
            case where there is no extractor structure.
    """

    wavelength: jnp.ndarray
    emitted_power: jnp.ndarray
    extracted_power: jnp.ndarray
    collected_power: jnp.ndarray
    bare_substrate_emitted_power: jnp.ndarray
    bare_substrate_extracted_power: jnp.ndarray
    bare_substrate_collected_power: jnp.ndarray


json_utils.register_custom_type(ExtractorResponse)

tree_util.register_pytree_node(
    ExtractorResponse,
    lambda r: (
        (
            r.wavelength,
            r.emitted_power,
            r.extracted_power,
            r.collected_power,
            r.bare_substrate_emitted_power,
            r.bare_substrate_extracted_power,
            r.bare_substrate_collected_power,
        ),
        None,
    ),
    lambda _, children: ExtractorResponse(*children),
)


class ExtractorComponent(base.Component):
    """Defines a photon extractor component."""

    def __init__(
        self,
        spec: ExtractorSpec,
        sim_params: ExtractorSimParams,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the photon extractor component.

        Args:
            spec: Defines the physical specification of the extractor.
            sim_params: Defines simulation parameters for the extractor.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.density_initializer = density_initializer

        _num_gridpoints = functools.partial(
            divide_and_round,
            b=spec.grid_spacing,
        )

        # When computing fields within each layer, a gridpoint is placed at the
        # very start and end of the layer, and so an additional gridpoint is needed
        # to ensure gridpoints are correctly spaced.
        self.layer_znum = (
            _num_gridpoints(spec.thickness_ambient) + 1,
            _num_gridpoints(spec.thickness_oxide) + 1,
            _num_gridpoints(spec.thickness_extractor) + 1,
            _num_gridpoints(spec.thickness_substrate_before_source) + 1,
            _num_gridpoints(spec.thickness_substrate_after_source) + 1,
        )

        self.seed_density = seed_density(
            grid_shape=self.spec.grid_shape,
            spec=self.spec,
            **seed_density_kwargs,
        )
        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.pitch * basis.X,
                v=self.spec.pitch * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the photon extractor component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        wavelength: Optional[float | jnp.ndarray] = None,
        expansion: Optional[basis.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[ExtractorResponse, base.AuxDict]:
        """Computes the response of the photon extractor.

        Args:
            params: The parameters defining the photon extractor, matching those
                returned by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, computes and xz cross section for electric
                and magnetic fields, which makes the calculation more expensive.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength

        return simulate_extractor(
            density=params,  # type: ignore[arg-type]
            spec=self.spec,
            layer_znum=self.layer_znum,
            wavelength=jnp.asarray(wavelength),
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )


def divide_and_round(a: float, b: float) -> int:
    """Checks that `a` is nearly evenly divisible by `b`, and returns `a / b`."""
    result = int(jnp.around(a / b))
    if not jnp.isclose(a / b, result):
        raise ValueError(
            f"`a` must be nearly evenly divisible by `b` spacing, but got `a` "
            f"{a} with `b` {b}."
        )
    return result


def seed_density(
    grid_shape: Tuple[int, int],
    spec: ExtractorSpec,
    **kwargs: Any,
) -> types.Density2DArray:
    """Return the seed density for a photon extractor component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        spec: Defines the physical structure of the photon extractor.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a photon extractor component.
    invalid_kwargs = (
        "array",
        "fixed_solid",
        "fixed_void",
        "lower_bound",
        "upper_bound",
        "periodic",
    )
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    design_mask = _mask(
        grid_shape,
        pitch=spec.pitch,
        width=spec.width_design_region,
    )
    fixed_void = ~design_mask

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        fixed_solid=jnp.zeros_like(fixed_void),
        fixed_void=fixed_void,
        periodic=(False, False),
        **kwargs,
    )


def simulate_extractor(
    density: types.Density2DArray,
    spec: ExtractorSpec,
    layer_znum: Tuple[int, int, int, int, int],
    wavelength: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
    compute_fields: bool,
) -> Tuple[ExtractorResponse, base.AuxDict]:
    """Simulates the photon extractor device.

    The response of the photon extractor includes the emitted, extracted, and collected
    power for x, y, and z-oriented dipoles beneath the extractor defined by `density`.
    It also includes the powers for dipoles in a bare substrate, i.e. lacking an
    extractor structure.

    Args:
        density: Defines the pattern of the photon extractor layer.
        spec: Defines the physical specifcation of the photon extractor.
        layer_znum: The number of gridpoints in the z-direction used for fields.
        wavelength: The wavelength of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.
        compute_fields: If `True`, returns electric and magnetic fields in the
            `aux` dictionary.

    Returns:
        The `ExtractorResponse` and `aux` dictionary.
    """
    density_array = utils.transforms.rescaled_density_array(
        density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    in_plane_wavevector = jnp.zeros((2,))
    primitive_lattice_vectors = basis.LatticeVectors(
        u=spec.pitch * basis.X,
        v=spec.pitch * basis.Y,
    )

    grid_shape: Tuple[int, int] = density_array.shape  # type: ignore[assignment]
    with jax.ensure_compile_time_eval():
        assert grid_shape == spec.grid_shape

    def eigensolve_pml(permittivity: jnp.ndarray) -> fmm.LayerSolveResult:
        # Permittivities and permeabilities are returned in the order needed
        # for the anisotropic eigensolve below.
        permittivities_pml, permeabilities_pml = pml.apply_uniaxial_pml(
            permittivity=permittivity,
            pml_params=_pml_params(grid_shape, spec),
        )
        return fmm.eigensolve_general_anisotropic_media(
            wavelength,
            in_plane_wavevector,
            primitive_lattice_vectors,
            *permittivities_pml,
            *permeabilities_pml,
            expansion=expansion,
            formulation=formulation,
            vector_field_source=jnp.mean(jnp.asarray(permittivities_pml), axis=0),
        )

    with jax.ensure_compile_time_eval():
        solve_result_ambient = eigensolve_pml(
            permittivity=jnp.full(grid_shape, spec.permittivity_ambient)
        )
        solve_result_substrate = eigensolve_pml(
            permittivity=jnp.full(grid_shape, spec.permittivity_substrate)
        )

    solve_result_oxide = eigensolve_pml(
        permittivity=utils.transforms.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_oxide),
            permittivity_void=jnp.asarray(spec.permittivity_ambient),
            density=density_array,
        ),
    )
    solve_result_extractor = eigensolve_pml(
        permittivity=utils.transforms.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_extractor),
            permittivity_void=jnp.asarray(spec.permittivity_ambient),
            density=density_array,
        ),
    )

    solve_results_before_source = (
        solve_result_ambient,
        solve_result_oxide,
        solve_result_extractor,
        solve_result_substrate,  # Before the source.
    )
    solve_results_after_source = (solve_result_substrate,)
    thicknesses_before_source = (
        jnp.asarray(spec.thickness_ambient),
        jnp.asarray(spec.thickness_oxide),
        jnp.asarray(spec.thickness_extractor),
        jnp.asarray(spec.thickness_substrate_before_source),
    )
    thicknesses_after_source = (jnp.asarray(spec.thickness_substrate_after_source),)

    # Compute scattering matrices for the structure above the source, which depends
    # upon the density array and cannot be done at compile time.
    if compute_fields:
        # If fields wanted, compute the full set of interior scattering matrices.
        s_matrices_interior_before_source = scattering.stack_s_matrices_interior(
            layer_solve_results=solve_results_before_source,
            layer_thicknesses=thicknesses_before_source,
        )
        s_matrix_before_source = s_matrices_interior_before_source[-1][0]
    else:
        s_matrix_before_source = scattering.stack_s_matrix(
            layer_solve_results=solve_results_before_source,
            layer_thicknesses=thicknesses_before_source,
        )

    # Scattering matrices for the structure below the source, and scattering matrices
    # for the bare substrate (i.e. no extractor structure) is done at compile time.
    with jax.ensure_compile_time_eval():
        if compute_fields:
            s_matrices_interior_after_source = scattering.stack_s_matrices_interior(
                layer_solve_results=solve_results_after_source,
                layer_thicknesses=thicknesses_after_source,
            )
            s_matrix_after_source = s_matrices_interior_after_source[-1][0]
        else:
            s_matrix_after_source = scattering.stack_s_matrix(
                layer_solve_results=solve_results_after_source,
                layer_thicknesses=thicknesses_after_source,
            )

        s_matrix_before_source_no_substrate = scattering.stack_s_matrix(
            layer_solve_results=(
                solve_result_ambient,  # ambient + oxide + extractor
                solve_result_substrate,
            ),
            layer_thicknesses=(
                jnp.asarray(
                    spec.thickness_ambient
                    + spec.thickness_oxide
                    + spec.thickness_extractor
                ),
                jnp.asarray(spec.thickness_substrate_before_source),
            ),
        )

        # Generate the Fourier representation of x, y, and z-oriented point dipoles.
        dipole = sources.gaussian_source(
            fwhm=jnp.asarray(spec.fwhm_source),
            location=jnp.asarray([[spec.pitch / 2, spec.pitch / 2]]),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            expansion=expansion,
        )
        zeros = jnp.zeros_like(dipole)
        jx = jnp.concatenate([dipole, zeros, zeros], axis=-1)
        jy = jnp.concatenate([zeros, dipole, zeros], axis=-1)
        jz = jnp.concatenate([zeros, zeros, dipole], axis=-1)

    def compute_power(
        s_matrix_before_source: scattering.ScatteringMatrix,
        s_matrix_after_source: scattering.ScatteringMatrix,
    ) -> Tuple[
        Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
        Tuple[jnp.ndarray, jnp.ndarray],
    ]:
        """Compute the emitted, extracted, and collected dipole power."""
        (
            bwd_amplitude_ambient_end,
            fwd_amplitude_before_start,
            bwd_amplitude_before_end,
            fwd_amplitude_after_start,
            bwd_amplitude_after_end,
            _,
        ) = sources.amplitudes_for_source(
            jx=jx,
            jy=jy,
            jz=jz,
            s_matrix_before_source=s_matrix_before_source,
            s_matrix_after_source=s_matrix_after_source,
        )

        # -------------------------------------------------------------------------
        # Total emitted power measured at monitors in the substrate.
        # -------------------------------------------------------------------------

        # Compute the Poynting flux in the layer before the source, at the monitor.
        fwd_amplitude_before_monitor = fields.propagate_amplitude(
            amplitude=fwd_amplitude_before_start,
            distance=jnp.asarray(
                spec.thickness_substrate_before_source - spec.offset_monitor_source
            ),
            layer_solve_result=solve_result_substrate,
        )
        bwd_amplitude_before_monitor = fields.propagate_amplitude(
            amplitude=bwd_amplitude_before_end,
            distance=jnp.asarray(spec.offset_monitor_source),
            layer_solve_result=solve_result_substrate,
        )
        (
            fwd_flux_before_monitor,
            bwd_flux_before_monitor,
        ) = fields.directional_poynting_flux(
            forward_amplitude=fwd_amplitude_before_monitor,
            backward_amplitude=bwd_amplitude_before_monitor,
            layer_solve_result=solve_result_substrate,
        )

        # Compute the Poynting flux in the layer after the source, at the monitor.
        fwd_amplitude_after_monitor = fields.propagate_amplitude(
            amplitude=fwd_amplitude_after_start,
            distance=jnp.asarray(spec.offset_monitor_source),
            layer_solve_result=solve_result_substrate,
        )
        bwd_amplitude_after_monitor = fields.propagate_amplitude(
            amplitude=bwd_amplitude_after_end,
            distance=jnp.asarray(
                spec.thickness_substrate_after_source - spec.offset_monitor_source
            ),
            layer_solve_result=solve_result_substrate,
        )
        (
            fwd_flux_after_monitor,
            bwd_flux_after_monitor,
        ) = fields.directional_poynting_flux(
            forward_amplitude=fwd_amplitude_after_monitor,
            backward_amplitude=bwd_amplitude_after_monitor,
            layer_solve_result=solve_result_substrate,
        )

        # Compute the total forward and backward flux resulting from the source. The
        # forward flux from the source is the difference between the forward flux just
        # after the source, and the forward flux just before the source. The backward
        # flux is defined analogously.
        fwd_flux_from_source = fwd_flux_after_monitor - fwd_flux_before_monitor
        bwd_flux_from_source = bwd_flux_before_monitor - bwd_flux_after_monitor

        # Sum the the flux over all Fourier orders.
        total_emitted = jnp.sum(fwd_flux_from_source, axis=-2) - jnp.sum(
            bwd_flux_from_source, axis=-2
        )

        # -------------------------------------------------------------------------
        # Total extracted power measured at a monitor above the extractor.
        # -------------------------------------------------------------------------

        with jax.ensure_compile_time_eval():
            print(s_matrix_before_source.start_layer_thickness)

        # Compute the eigenmode amplitudes at the ambient flux monitor.
        bwd_amplitude_ambient_monitor = fields.propagate_amplitude(
            amplitude=bwd_amplitude_ambient_end,
            distance=jnp.asarray(
                s_matrix_before_source.start_layer_thickness
                - (spec.thickness_ambient - spec.offset_monitor_ambient)
            ),
            layer_solve_result=solve_result_ambient,
        )
        _, bwd_flux_ambient_monitor = fields.directional_poynting_flux(
            forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_monitor),
            backward_amplitude=bwd_amplitude_ambient_monitor,
            layer_solve_result=solve_result_ambient,
        )
        total_extracted = -jnp.sum(bwd_flux_ambient_monitor, axis=-2)

        # We also want to compute the power collected by a monitor that is located above
        # the extractor design, and does not extend to the edges of the unit cell. To
        # find the flux through this monitor, compute the flux on the real-space grid
        # and sum over the target region.
        #
        # First compute Fourier amplitudes of the electric and magnetic fields.
        ambient_monitor_ef, ambient_monitor_hf = fields.fields_from_wave_amplitudes(
            jnp.zeros_like(bwd_amplitude_ambient_monitor),
            bwd_amplitude_ambient_monitor,
            layer_solve_result=solve_result_ambient,
        )
        # Compute the real-space electric and magnetic fields at the monitor.
        ambient_monitor_ef, ambient_monitor_hf, (x, y) = fields.fields_on_grid(
            electric_field=ambient_monitor_ef,
            magnetic_field=ambient_monitor_hf,
            layer_solve_result=solve_result_ambient,
            shape=grid_shape,
            num_unit_cells=(1, 1),
        )
        assert ambient_monitor_ef[0].shape == wavelength.shape + grid_shape + (3,)
        # Compute the Poynting flux on the real-space grid at the monitor.
        bwd_flux_ambient_monitor = _time_average_z_poynting_flux(
            electric_field=ambient_monitor_ef,
            magnetic_field=ambient_monitor_hf,
        )
        # Compute the masked flux.
        monitor_mask = _mask(
            grid_shape=grid_shape,
            pitch=spec.pitch,
            width=spec.width_monitor_ambient,
        )
        masked_bwd_flux_ambient_monitor = jnp.where(
            monitor_mask[..., jnp.newaxis],
            bwd_flux_ambient_monitor,
            0.0,
        )
        total_collected = -jnp.mean(masked_bwd_flux_ambient_monitor, axis=(-3, -2))
        assert total_extracted.shape == total_emitted.shape == total_collected.shape

        return (
            (total_emitted, total_extracted, total_collected),
            (bwd_amplitude_before_end, fwd_amplitude_after_start),
        )

    # Compute the emitted, extracted, and collected power for the photon extractor.
    (total_emitted, total_extracted, total_collected), (
        bwd_amplitude_before_end,
        fwd_amplitude_after_start,
    ) = compute_power(
        s_matrix_before_source=s_matrix_before_source,
        s_matrix_after_source=s_matrix_after_source,
    )

    # Compute the emitted, extracted, and collected power for the bare substrate.
    with jax.ensure_compile_time_eval():
        (
            bare_substrate_total_emitted,
            bare_substrate_total_extracted,
            bare_substrate_total_collected,
        ), _ = compute_power(
            s_matrix_before_source=s_matrix_before_source_no_substrate,
            s_matrix_after_source=s_matrix_after_source,
        )

    response = ExtractorResponse(
        wavelength=wavelength,
        emitted_power=total_emitted,
        extracted_power=total_extracted,
        collected_power=total_collected,
        bare_substrate_emitted_power=bare_substrate_total_emitted,
        bare_substrate_extracted_power=bare_substrate_total_extracted,
        bare_substrate_collected_power=bare_substrate_total_collected,
    )

    # -------------------------------------------------------------------------
    # Compute fields in an xz cross section.
    # -------------------------------------------------------------------------

    aux = {}
    if compute_fields:
        amplitudes_interior = fields.stack_amplitudes_interior_with_source(
            s_matrices_interior_before_source=s_matrices_interior_before_source,
            s_matrices_interior_after_source=s_matrices_interior_after_source,
            backward_amplitude_before_end=bwd_amplitude_before_end,
            forward_amplitude_after_start=fwd_amplitude_after_start,
        )
        x = jnp.linspace(0, spec.pitch, grid_shape[0])
        y = jnp.ones_like(x) * spec.pitch / 2
        (ex, ey, ez), (hx, hy, hz), (x, y, z) = fields.stack_fields_3d_on_coordinates(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=(
                solve_results_before_source + solve_results_after_source
            ),
            layer_thicknesses=thicknesses_before_source + thicknesses_after_source,
            layer_znum=layer_znum,
            x=x,
            y=y,
        )
        aux.update(
            {
                EFIELD: (ex, ey, ez),
                HFIELD: (hx, hy, hz),
                FIELD_COORDINATES: (x, y, z),
            }
        )

    return response, aux


def _pml_params(grid_shape: Tuple[int, int], spec: ExtractorSpec) -> pml.PMLParams:
    """Return PML parameters for the specified grid shape and extractor spec."""
    return pml.PMLParams(
        num_x=int(grid_shape[0] * spec.width_pml / spec.pitch),
        num_y=int(grid_shape[1] * spec.width_pml / spec.pitch),
    )


def _time_average_z_poynting_flux(
    electric_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """Computes the time-average z-directed Poynting flux given the physical fields."""
    # https://github.com/facebookresearch/fmmax/blob/main/examples/sorter.py
    ex, ey, _ = electric_field
    hx, hy, _ = magnetic_field
    return jnp.real(ex * jnp.conj(hy) - ey * jnp.conj(hx))


def _mask(
    grid_shape: Tuple[int, int],
    pitch: float,
    width: float,
) -> jnp.ndarray:
    """Generate a mask that is `True` in a centered region having width `width`"""
    x, y = jnp.meshgrid(
        jnp.arange(0.5, grid_shape[0]) * pitch / grid_shape[0],
        jnp.arange(0.5, grid_shape[1]) * pitch / grid_shape[1],
        indexing="ij",
    )
    x_offset = (pitch - width) / 2
    y_offset = (pitch - width) / 2
    return (
        (x >= x_offset)
        & (y >= y_offset)
        & (x < pitch - x_offset)
        & (y < pitch - y_offset)
    )



"""Defines the meta-atom library component and simulation routine.

Copyright (c) 2024 The INVRS-IO authors.
"""

import dataclasses
import functools
import pathlib
from typing import Any, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as onp
from fmmax import basis, fields, fmm, scattering
from jax import tree_util
from totypes import json_utils, types

from invrs_gym.challenges import base
from invrs_gym.utils import materials, transforms

Params = Dict[str, types.BoundedArray | types.Density2DArray | jnp.ndarray]

DENSITY = "density"
THICKNESS = "thickness"

EFIELD_XZ = "efield_xz"
HFIELD_XZ = "hfield_xz"
COORDS_XZ = "coordinates_xz"

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0

# Register a material with the refractive index data for TiO2 taken from Chen et al.
# https://www.nature.com/articles/s41467-023-38185-2
TIO2_CHEN = "tio2_chen"
TIO2_DATA_PATH = pathlib.Path(__file__).resolve().parent / "data/nk_tio2.csv"
materials.register_material(TIO2_CHEN, TIO2_DATA_PATH)


@dataclasses.dataclass
class LibrarySpec:
    """Defines the physical specification of the meta-atom library.

    Attributes:
        material_ambient: The ambient material.
        material_metasurface_solid: The material comprising metasurface solid regions.
        material_metasurface_void: The material comprising metasurface void regions.
        material_substrate: The substrate material.
        background_extinction_coefficient: Additional extinction coefficient to be
            included in the permittivity. Positive values correspond to optical loss
            and help stabilize the calculation.
        thickness_ambient: The thickness of the ambient layer.
        thickness_metasurface: The thickness of the metasurface layer.
        thickness_substrate: The thickness of the substrate layer.
        pitch: The pitch of the meta-atom, i.e. spacing in x and y directions.
        frame_width: The width of the frame surrounding each meta-atom. In the frame,
            all pixels are required to be void.
        grid_spacing: The spacing of the grid on which meta-atom density is defined.
    """

    material_ambient: materials.Material
    material_metasurface_solid: materials.Material
    material_metasurface_void: materials.Material
    material_substrate: materials.Material

    background_extinction_coeff: float

    thickness_ambient: float
    thickness_metasurface: float | jnp.ndarray | types.BoundedArray
    thickness_substrate: float

    pitch: float
    frame_width: float

    grid_spacing: float

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        with jax.ensure_compile_time_eval():
            return (
                int(jnp.ceil(self.pitch / self.grid_spacing).astype(int)),
                int(jnp.ceil(self.pitch / self.grid_spacing).astype(int)),
            )

    @property
    def frame_pixels(self) -> int:
        """Return the width of the frame in pixels."""
        with jax.ensure_compile_time_eval():
            return int(jnp.ceil(self.frame_width / self.grid_spacing))


@dataclasses.dataclass
class LibrarySimParams:
    """Stores simulation parameters for the meta-atom library.

    Attributes:
        wavelength: The wavelength of the excitation.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        formulation: The FMM formulation used for the calculation.
        truncation: The truncation of lattice vectors for the calculation.
    """

    wavelength: jnp.ndarray
    approximate_num_terms: int
    formulation: fmm.Formulation
    truncation: basis.Truncation


@dataclasses.dataclass
class LibraryResponse:
    """Stores the response of meta-atoms from the library.

    Attributes:
        wavelength: The wavelength of the excitation.
        transmission_rhcp: The complex amplitude transmission for excitation with
            right-hand circularly-polarized light. The coefficients are for right-
            hand and left-hand circular polarizations, i.e. the conserved and
            converted polarizations.
        transmission_lhcp: The complex amplitude transmission for excitation with
            left-hand circularly-polarized light.
        reflection_rhcp: The complex amplitude reflection for excitation with right-
            hand circularly-polarized light.
        reflection_lhcp: The complex amplitude reflection for excitation with left-
            hand circularly-polarized light.
    """

    wavelength: jnp.ndarray
    transmission_rhcp: jnp.ndarray
    transmission_lhcp: jnp.ndarray
    reflection_rhcp: jnp.ndarray
    reflection_lhcp: jnp.ndarray


tree_util.register_dataclass(
    nodetype=LibraryResponse,
    data_fields=[
        "wavelength",
        "transmission_rhcp",
        "transmission_lhcp",
        "reflection_rhcp",
        "reflection_lhcp",
    ],
    meta_fields=[],
)

json_utils.register_custom_type(LibraryResponse)


class LibraryComponent(base.Component):
    """Defines the meta-atom library component."""

    def __init__(
        self,
        spec: LibrarySpec,
        sim_params: LibrarySimParams,
        library_size: int,
        thickness_initializer: base.ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs,
    ) -> None:
        """Initializes the meta-atom library component.

        Args:
            spec: The physical specification for the library of meta-atoms.
            sim_params: The default simulation parameters for the library of meta-atoms.
            library_size: The size of the library.
            thickness_initializer: Callable which generates the initial metasurface
                thickness from a random key and the seed thickness.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """
        self.spec = spec
        self.sim_params = sim_params
        self.library_size = library_size
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_density = seed_density(
            library_size=library_size,
            grid_shape=spec.grid_shape,
            frame_pixels=spec.frame_pixels,
            **seed_density_kwargs,
        )
        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=spec.pitch * basis.X, v=spec.pitch * basis.Y
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for a meta-atom library."""
        key_thickness, key_density = jax.random.split(key)
        params = {
            THICKNESS: self.thickness_initializer(
                key_thickness, self.spec.thickness_metasurface  # type: ignore[arg-type]
            ),
            DENSITY: self.density_initializer(key_density, self.seed_density),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        wavelength: Optional[jnp.ndarray] = None,
        expansion: Optional[basis.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[LibraryResponse, base.AuxDict]:
        """Computes the response of the meta-atom library."""
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        if expansion is None:
            expansion = self.expansion

        spec = dataclasses.replace(
            self.spec,
            thickness_metasurface=(
                params[THICKNESS].array,  # type: ignore[arg-type, union-attr]
            ),
        )

        return simulate_library(
            density=params[DENSITY],  # type: ignore[arg-type]
            spec=spec,
            wavelength=wavelength,
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )


def seed_density(
    library_size: int,
    grid_shape: Tuple[int, int],
    frame_pixels: int,
    **kwargs: Any,
) -> types.Density2DArray:
    """Return the seed density for a meta-atom library.

    Args:
        library_size: The number of meta-atoms in the library. Determines the batch
            size of the density.
        grid_shape: The shape of the grid on which the density is defined.
        frame_pixels: The size of the fixed void frame surrounding each meta atom, in
            pixels.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a sorter component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic", "fixed_void")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    fixed_void = onp.zeros(grid_shape, dtype=bool)
    fixed_void[:frame_pixels, :] = True
    fixed_void[:, :frame_pixels] = True
    fixed_void[grid_shape[0] - frame_pixels :, :] = True
    fixed_void[:, grid_shape[1] - frame_pixels :] = True
    return types.Density2DArray(
        array=jnp.full((library_size,) + grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(False, False),
        fixed_solid=None,
        fixed_void=fixed_void,
        **kwargs,
    )


def simulate_library(
    density: types.Density2DArray,
    spec: LibrarySpec,
    wavelength: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
    compute_fields: bool,
) -> Tuple[LibraryResponse, base.AuxDict]:
    """Compute zeroth order complex reflection and transmission for each unit cell."""
    density_array = transforms.rescaled_density_array(
        density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    assert density_array.ndim in (2, 3)
    if density_array.ndim == 3:
        # Shift the batch axis of the density so it lies ahead of other batch axes.
        batch_dims = wavelength.ndim
        density_array = density_array.reshape(
            density_array.shape[0], *([1] * batch_dims), *density_array.shape[1:]
        )

    with jax.ensure_compile_time_eval():
        assert density_array.shape[-2] % spec.grid_shape[0] == 0
        assert density_array.shape[-1] % spec.grid_shape[1] == 0
    period_x = density_array.shape[-2] // spec.grid_shape[0] * spec.pitch
    period_y = density_array.shape[-1] // spec.grid_shape[1] * spec.pitch

    in_plane_wavevector = jnp.zeros((2,))
    eigensolve = functools.partial(
        fmm.eigensolve_isotropic_media,
        wavelength=jnp.asarray(wavelength),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=basis.LatticeVectors(
            u=period_x * basis.X,
            v=period_y * basis.Y,
        ),
        expansion=expansion,
        formulation=formulation,
    )

    def permittivity_fn(material: materials.Material) -> jnp.ndarray:
        dims_to_add = 1 + batch_dims - wavelength.ndim
        shape = (1,) * dims_to_add + wavelength.shape + (1, 1)
        return materials.permittivity(
            material=material,
            wavelength_um=wavelength,
            background_extinction_coeff=spec.background_extinction_coeff,
        ).reshape(shape)

    # Perform eigensolve for all the non-metasurface layers.
    with jax.ensure_compile_time_eval():
        solve_result_ambient = eigensolve(
            permittivity=permittivity_fn(spec.material_ambient)
        )
        solve_result_substrate = eigensolve(
            permittivity=permittivity_fn(spec.material_substrate)
        )

    permittivity_metasurface = transforms.interpolate_permittivity(
        permittivity_solid=permittivity_fn(spec.material_metasurface_solid),
        permittivity_void=permittivity_fn(spec.material_metasurface_void),
        density=density_array,
    )

    solve_result_metasurface = eigensolve(permittivity=permittivity_metasurface)
    layer_solve_results = (
        solve_result_ambient,
        solve_result_metasurface,
        solve_result_substrate,
    )
    layer_thicknesses = (
        jnp.asarray(spec.thickness_ambient),
        jnp.asarray(spec.thickness_metasurface),
        jnp.asarray(spec.thickness_substrate),
    )

    if compute_fields:
        s_matrices_interior = scattering.stack_s_matrices_interior(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )
        s_matrix = s_matrices_interior[-1][0]
    else:
        s_matrix = scattering.stack_s_matrix(layer_solve_results, layer_thicknesses)

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate the incident wave amplitudes for two linear polarizations. The first
    # has electric field oriented in the x direction, the second has electric field
    # oriented in the y direction. Note that the power is incident from the substrate,
    # i.e. it is travelling "backward" in the fmmax convention.
    incident = jnp.zeros((2 * n, 2), dtype=complex)
    incident = incident.at[n, 0].set(1.0)
    incident = incident.at[0, 1].set(1.0)

    # Normalize the incident amplitudes, dividing by the square root of incident power.
    _, incident_power = fields.amplitude_poynting_flux(
        forward_amplitude=jnp.zeros_like(incident),
        backward_amplitude=incident,
        layer_solve_result=solve_result_substrate,
    )
    total_incident_power = jnp.sum(incident_power, axis=-2, keepdims=True)
    incident /= jnp.sqrt(jnp.abs(total_incident_power))

    # Calculate the reflected and transmitted mode amplitudes.
    reflected = s_matrix.s12 @ incident
    transmitted = s_matrix.s22 @ incident

    # Create the reflection matrix. It multiplies incident ampltudes in the linear
    # basis to give reflected zeroth-order fields in the linear basis.
    r_ex_from_ex = reflected[..., n, 0] / incident[..., n, 0]
    r_ey_from_ex = reflected[..., 0, 0] / incident[..., n, 0]
    r_ex_from_ey = reflected[..., n, 1] / incident[..., 0, 1]
    r_ey_from_ey = reflected[..., 0, 1] / incident[..., 0, 1]
    reflection_linear = jnp.stack(
        [
            jnp.stack([r_ex_from_ex, r_ex_from_ey], axis=-1),
            jnp.stack([r_ey_from_ex, r_ey_from_ey], axis=-1),
        ],
        axis=-2,
    )
    reflection_circular = circular_from_linear(reflection_linear)

    # Same operations for transmission coefficients.
    t_ex_from_ex = transmitted[..., n, 0] / incident[..., n, 0]
    t_ey_from_ex = transmitted[..., 0, 0] / incident[..., n, 0]
    t_ex_from_ey = transmitted[..., n, 1] / incident[..., 0, 1]
    t_ey_from_ey = transmitted[..., 0, 1] / incident[..., 0, 1]
    transmission_linear = jnp.stack(
        [
            jnp.stack([t_ex_from_ex, t_ex_from_ey], axis=-1),
            jnp.stack([t_ey_from_ex, t_ey_from_ey], axis=-1),
        ],
        axis=-2,
    )
    transmission_circular = circular_from_linear(transmission_linear)

    response = LibraryResponse(
        wavelength=wavelength,
        transmission_rhcp=transmission_circular[..., 0],
        transmission_lhcp=transmission_circular[..., 1],
        reflection_rhcp=reflection_circular[..., 0],
        reflection_lhcp=reflection_circular[..., 1],
    )

    aux = {}
    if compute_fields:
        amplitudes_interior = fields.stack_amplitudes_interior(
            s_matrices_interior=s_matrices_interior,
            forward_amplitude_0_start=jnp.zeros_like(incident),
            backward_amplitude_N_end=incident,
        )
        layer_znum = tuple(
            [int(jnp.round(t / spec.grid_spacing) + 1) for t in layer_thicknesses]
        )
        x = jnp.arange(density_array.shape[-2]) * spec.grid_spacing
        y = jnp.full_like(x, spec.pitch / 2)
        efield_xz, hfield_xz, coords_xz = fields.stack_fields_3d_on_coordinates(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
            layer_znum=layer_znum,
            x=x,
            y=y,
        )
        aux.update({EFIELD_XZ: efield_xz, HFIELD_XZ: hfield_xz, COORDS_XZ: coords_xz})

    return response, aux


def circular_from_linear(x: jnp.ndarray) -> jnp.ndarray:
    """Return circular reflection or transmission coefficients from linear."""
    inv_conversion_matrix = jnp.asarray([[0.5, 0.5j], [0.5, -0.5j]])
    conversion_matrix = jnp.linalg.inv(inv_conversion_matrix)
    return inv_conversion_matrix @ x @ conversion_matrix



"""Defines functions that compute material properties.

Copyright (c) 2024 The INVRS-IO authors.
"""

import functools
import pathlib
import warnings
from typing import Dict, Protocol, Union

import jax
import numpy as onp
import refractiveindex as ri  # type: ignore[import-untyped]
from jax import numpy as jnp

Material = str | ri.RefractiveIndexMaterial

# Define some common materials.
SI = ri.RefractiveIndexMaterial(shelf="main", book="Si", page="Green-1995")
AMORPHOUS_SI = ri.RefractiveIndexMaterial(shelf="main", book="Si", page="Pierce")
SIO2 = ri.RefractiveIndexMaterial(shelf="main", book="SiO2", page="Malitson")
SI3N4 = ri.RefractiveIndexMaterial(shelf="main", book="Si3N4", page="Luke")
TIO2 = ri.RefractiveIndexMaterial(shelf="main", book="TiO2", page="Jolivet-amorphous")
VACUUM = "vacuum"  # Permittivity is computed via dedicated function.


def permittivity(
    material: str | ri.RefractiveIndexMaterial,
    wavelength_um: jnp.ndarray,
    background_extinction_coeff: float = 0.0,
) -> jnp.ndarray:
    """Return the permittivity for the specified material.

    Args:
        material: The name of a registered material, or a `ri.RefractiveIndexMaterial`
            for which the permittivity is sought.
        wavelength_um: The wavelength for the permittivity, in units of microns.
        background_extinction_coeff: Additional extinction coefficient to be included
            in the permittivity. Positive values correspond to optical loss, and the
            default value is 0.

    Returns:
        The permittivity of the material at the specified wavelengths.
    """
    if isinstance(material, ri.RefractiveIndexMaterial):
        permittivity_fn: PermittivityFn = functools.partial(
            permittivity_from_database,
            material=material,
        )
    else:
        permittivity_fn = PERMITTIVITY_FNS[material]
    return permittivity_fn(
        wavelength_um=wavelength_um,
        background_extinction_coeff=background_extinction_coeff,
    )


def permittivity_from_database(
    material: ri.RefractiveIndexMaterial,
    wavelength_um: jnp.ndarray,
    background_extinction_coeff: float,
) -> jnp.ndarray:
    """Return the permittivity for the specified material from the database."""

    def _jax_fn(wavelength_um: jnp.ndarray) -> jnp.ndarray:
        wavelength_nm = 1000 * wavelength_um
        try:
            epsilon = material.get_epsilon(wavelength_nm)
            refractive_index = onp.sqrt(epsilon)
        except ri.refractiveindex.NoExtinctionCoefficient:
            refractive_index = material.get_refractive_index(wavelength_nm)
        epsilon = (refractive_index + 1j * background_extinction_coeff) ** 2
        return jnp.asarray(epsilon, dtype=jnp.zeros((), dtype=complex).dtype)

    result_shape_dtypes = jnp.zeros_like(wavelength_um, dtype=complex)
    return jax.pure_callback(_jax_fn, result_shape_dtypes, wavelength_um)


def permittivity_vacuum(
    wavelength_um: jnp.ndarray,
    background_extinction_coeff: float = 0.0,
) -> jnp.ndarray:
    """Return the permittivity of vacuum, with optional background extinction coeff."""
    return jnp.full(wavelength_um.shape, 1.0 + 1j * background_extinction_coeff)


class PermittivityFn(Protocol):
    def __call__(
        self, wavelength_um: jnp.ndarray, background_extinction_coeff: float
    ) -> jnp.ndarray:
        ...


PERMITTIVITY_FNS: Dict[str, PermittivityFn] = {
    VACUUM: permittivity_vacuum,
}


def register_material(name: str, path: Union[str, pathlib.Path]) -> None:
    """Registers a material so it is accessible by `get_permittivity` function."""
    with jax.ensure_compile_time_eval():
        data_wvl_n_k = onp.genfromtxt(path, delimiter=",", comments="#")
        data_wavelength_um = jnp.asarray(data_wvl_n_k[:, 0])
        data_permittivity = (data_wvl_n_k[:, 1] + 1j * data_wvl_n_k[:, 2]) ** 2

        if name in PERMITTIVITY_FNS:
            warnings.warn(f"Material {name} already registered.")

        def _permittivity_fn(
            wavelength_um: jnp.ndarray,
            background_extinction_coeff: float,
        ) -> jnp.ndarray:
            refractive_index = jnp.sqrt(
                jnp.interp(wavelength_um, data_wavelength_um, data_permittivity)
            )
            return (refractive_index + 1j * background_extinction_coeff) ** 2

        PERMITTIVITY_FNS[name] = _permittivity_fn



"""Defines the diffractive beamsplitter challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
import itertools
from typing import Dict, Sequence, Tuple

import jax.numpy as jnp
from fmmax import basis, fmm  # type: ignore[import-untyped]
from totypes import types

from invrs_gym import utils
from invrs_gym.challenges import base
from invrs_gym.challenges.diffract import common
from invrs_gym.loss import transmission_loss

Params = Dict[str, types.BoundedArray | types.Density2DArray]


DENSITY = "density"
THICKNESS = "thickness"

TOTAL_EFFICIENCY = "total_efficiency"
AVERAGE_EFFICIENCY = "average_efficiency"
MIN_EFFICIENCY = "min_efficiency"
ZEROTH_ORDER_EFFICIENCY = "zeroth_order_efficiency"
ZEROTH_ORDER_ERROR = "zeroth_order_error"
UNIFORMITY_ERROR = "uniformity_error"
UNIFORMITY_ERROR_WITHOUT_ZEROTH_ORDER = "uniformity_error_without_zeroth_order"

POLARIZATION = "TE"

TRANSMISSION_EXPONENT = 1.0
SCALAR_EXPONENT = 2.0

density_initializer = functools.partial(
    utils.initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class DiffractiveSplitterChallenge(base.Challenge):
    """Defines the diffractive beamsplitter challenge.

    The objective of the challenge is to evenly split light into an array of
    transmitted diffraction orders. The challenge is based on the LightTrans
    publication, "Design and rigorous analysis of a non-paraxial diffractive
    beamsplitter", retrieved from
    https://www.lighttrans.com/fileadmin/shared/UseCases/Application_UC_Rigorous%20Analysis%20of%20Non-paraxial%20Diffractive%20Beam%20Splitter.pdf

    Attributes:
        component: The component to be designed.
        splitting: Defines the target splitting for the beamsplitter.
    """

    component: common.GratingWithOptimizableThicknessComponent
    splitting: Tuple[int, int]

    def loss(
        self,
        response: common.GratingResponse,
        normalized_efficiency_lower_bound: float = 0.6,
        normalized_efficiency_upper_bound: float = 0.8,
    ) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`.

        Args:
            response: The response obtained from simulating the component.
            normalized_efficiency_lower_bound: Value used in a nonlinearity within the
                loss function. Together with `normalized_efficiency_upper_bound`,
                defines a window within which the loss is less sensitive to efficiency,
                which promotes convergence to values within the window.
            normalized_efficiency_upper_bound: Defines the upper bound of the window.

        Returns:
            The scalar loss.
        """
        efficiency = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
            polarization=POLARIZATION,
        )
        batch_shape = jnp.broadcast_shapes(
            response.wavelength.shape,
            response.polar_angle.shape,
            response.azimuthal_angle.shape,
        )
        assert efficiency.shape == batch_shape + self.splitting + (1,)
        num_splits = self.splitting[0] * self.splitting[1]
        lower_bound = normalized_efficiency_lower_bound / num_splits
        upper_bound = normalized_efficiency_upper_bound / num_splits

        # Compute per-wavelength orthotope loss.
        loss = transmission_loss.orthotope_smooth_transmission_loss(
            transmission=efficiency,
            window_lower_bound=jnp.full(efficiency.shape, lower_bound),
            window_upper_bound=jnp.full(efficiency.shape, upper_bound),
            transmission_exponent=jnp.asarray(TRANSMISSION_EXPONENT),
            scalar_exponent=jnp.asarray(SCALAR_EXPONENT),
            axis=(-3, -2, -1),
        )
        return jnp.mean(loss)  # Mean reduction across wavelengths, if they exist.

    def eval_metric(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute eval metric from the component `response`.

        The eval metric rewards high total efficiency and minimum nonuniformity.
        It is computed by,

            eval_metric = total_efficiency * (1 - uniformity_error)

        In cases where multiple wavelengths or incident angles are considered, the
        eval metric is the minimum across all excitation conditions.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        transmission = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
            polarization=POLARIZATION,
        )
        assert transmission.shape[-3:] == self.splitting + (1,)

        # Total efficiency, i.e. sum of power into all output orders. A perfect
        # solution has a value of `1`, and the lowest possible value is `0`.
        total_efficiency = jnp.sum(transmission, axis=(-3, -2, -1))

        # Uniformity error. The perfect solution has a uniformity error of `0`,
        # and the highest possible value is `1`.
        uniformity_error = (
            jnp.amax(transmission, axis=(-3, -2, -1))
            - jnp.amin(transmission, axis=(-3, -2, -1))
        ) / (
            jnp.amax(transmission, axis=(-3, -2, -1))
            + jnp.amin(transmission, axis=(-3, -2, -1))
        )

        return jnp.amin(total_efficiency * (1 - uniformity_error))

    def metrics(
        self,
        response: common.GratingResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Args:
            response: The response of the diffractive splitter component.
            params: The parameters where the response was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary, with the following quantities:
                - total efficiency
                - average efficiency
                - zeroth order efficiency
                - zeroth order error
                - uniformity error
                - uniformity error without zeroth order
            For details, see slide 14 of the LightTrans publication.
        """
        metrics = super().metrics(response, params, aux)
        transmission = extract_orders_for_splitting(
            response.transmission_efficiency,
            expansion=response.expansion,
            splitting=self.splitting,
            polarization=POLARIZATION,
        )
        # Metrics are averaged over the batch axes, if they exist.
        total_efficiency = jnp.mean(jnp.sum(transmission, axis=(-3, -2, -1)))
        average_efficiency = jnp.mean(transmission)
        min_efficiency = jnp.amin(transmission)

        i0 = self.splitting[0] // 2
        j0 = self.splitting[1] // 2
        zeroth_efficiency = jnp.mean(transmission[..., i0, j0, :])
        zeroth_error = (zeroth_efficiency - average_efficiency) / average_efficiency

        uniformity_error = (jnp.amax(transmission) - jnp.amin(transmission)) / (
            jnp.amax(transmission) + jnp.amin(transmission)
        )

        mask = jnp.ones(transmission.shape, dtype=bool)
        mask = mask.at[..., i0, j0, :].set(False)
        masked_max_transmission = jnp.amax(jnp.where(mask, transmission, 0.0))
        masked_min_transmission = jnp.amin(jnp.where(mask, transmission, 1.0))

        uniformity_error_without_zeroth = (
            masked_max_transmission - masked_min_transmission
        ) / (masked_max_transmission + masked_min_transmission)

        metrics.update(
            {
                TOTAL_EFFICIENCY: total_efficiency,
                AVERAGE_EFFICIENCY: average_efficiency,
                MIN_EFFICIENCY: min_efficiency,
                ZEROTH_ORDER_EFFICIENCY: zeroth_efficiency,
                ZEROTH_ORDER_ERROR: zeroth_error,
                UNIFORMITY_ERROR: uniformity_error,
                UNIFORMITY_ERROR_WITHOUT_ZEROTH_ORDER: uniformity_error_without_zeroth,
            }
        )
        return metrics


def indices_for_splitting(
    expansion: basis.Expansion,
    splitting: Tuple[int, int],
) -> Tuple[int, ...]:
    """Return the indices for the orders identified by the target `splitting`."""
    idxs = []
    orders_x = range(-splitting[0] // 2 + 1, splitting[0] // 2 + 1)
    orders_y = range(-splitting[1] // 2 + 1, splitting[1] // 2 + 1)
    for ox, oy in itertools.product(orders_x, orders_y):
        idxs.append(common.index_for_order((ox, oy), expansion))
    return tuple(idxs)


def extract_orders_for_splitting(
    array: jnp.ndarray,
    expansion: basis.Expansion,
    splitting: Tuple[int, int],
    polarization: str,
) -> jnp.ndarray:
    """Extract the values from `array` for the specified splitting."""
    assert polarization in ("TE", "TM")
    polarization_idx = 0 if polarization == "TE" else 1
    array = array[..., polarization_idx, jnp.newaxis]

    num_splits = splitting[0] * splitting[1]
    shape = array.shape[:-2] + splitting + (array.shape[-1],)
    flat_shape = array.shape[:-2] + (num_splits, array.shape[-1])

    idxs = indices_for_splitting(expansion, splitting)
    extracted = jnp.zeros(flat_shape, dtype=array.dtype)
    extracted = extracted.at[..., :, :].set(array[..., idxs, :])
    return extracted.reshape(shape)


# -----------------------------------------------------------------------------
# Diffractive splitter with 7.2 x 7.2 um design region.
# -----------------------------------------------------------------------------


DIFFRACTIVE_SPLITTER_SPEC = common.GratingSpec(
    permittivity_ambient=(1.46 + 0.0j) ** 2,
    permittivity_cap=(1.46 + 0.0j) ** 2,
    # Small imaginary part stabilizes the FMM calculation.
    permittivity_grating=(1.46 + 0.00001j) ** 2,
    permittivity_encapsulation=(1.0 + 0.00001j) ** 2,
    permittivity_spacer=(1.0 + 0.0j) ** 2,
    permittivity_substrate=(1.0 + 0.0j) ** 2,
    thickness_cap=types.BoundedArray(array=0.0, lower_bound=0.0, upper_bound=0.1),
    thickness_grating=types.BoundedArray(array=0.692, lower_bound=0.5, upper_bound=1.5),
    thickness_spacer=types.BoundedArray(array=0.0, lower_bound=0.0, upper_bound=0.1),
    period_x=7.2,
    period_y=7.2,
    grid_spacing=0.04,  # Yields a grid shape of `(180, 180)`.
)

DIFFRACTIVE_SPLITTER_SIM_PARAMS = common.GratingSimParams(
    wavelength=0.6328,
    polar_angle=0.0,
    azimuthal_angle=0.0,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    approximate_num_terms=800,
    truncation=basis.Truncation.CIRCULAR,
)

# Objective is to split into a 7 x 7 array of beams.
SPLITTING = (7, 7)


def diffractive_splitter(
    minimum_width: int = 10,
    minimum_spacing: int = 10,
    thickness_initializer: base.ThicknessInitializer = (
        utils.initializers.identity_initializer
    ),
    density_initializer: base.DensityInitializer = density_initializer,
    splitting: Tuple[int, int] = SPLITTING,
    spec: common.GratingSpec = DIFFRACTIVE_SPLITTER_SPEC,
    sim_params: common.GratingSimParams = DIFFRACTIVE_SPLITTER_SIM_PARAMS,
    symmetries: Sequence[str] = (),
) -> DiffractiveSplitterChallenge:
    """Non-paraxial diffractive beamsplitter challenge.

    The diffractive splitter is based on "Design and rigorous analysis of a
    non-paraxial diffractive beamsplitter", an example of the LightTrans software
    (https://www.lighttrans.com/use-cases/application/design-and-rigorous-analysis-of-non-paraxial-diffractive-beam-splitter.html).

    It involves splitting a normally-incident TM-polarized plane wave into an
    array of 7x7 beams with maximal efficiency and uniformity.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            default value of 10 corresponds to a physical size of approximately 400 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        thickness_initializer: Callable which returns the initial thickness, given a
            key and seed thickness.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.
        splitting: Defines shape of the beam array to be created by the splitter.
        spec: Defines the physical specification of the diffractive splitter.
        sim_params: Defines the simulation settings of the diffractive splitter.
        symmetries: Defines the symmetries of the diffractive splitter.

    Returns:
        The `DiffractiveSplitterChallenge`.
    """
    return DiffractiveSplitterChallenge(
        component=common.GratingWithOptimizableThicknessComponent(
            spec=spec,
            sim_params=sim_params,
            thickness_initializer=thickness_initializer,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        splitting=splitting,
    )
